@@ -5,7 +5,7 @@ import errno
 import os
 import socket
 import sys
-from tests import skipped, LimitedTestCase, skip_with_pyevent
+from tests import skipped, LimitedTestCase, skip_with_pyevent, skip_if_no_ssl
 from unittest import main
 
 from eventlet import greenio
@@ -124,8 +124,8 @@ def read_http(sock):
             raise ConnectionClosed
         raise
     if not response_line:
-        raise ConnectionClosed
-    
+        raise ConnectionClosed(response_line)
+
     header_lines = []
     while True:
         line = fd.readline()
@@ -151,6 +151,7 @@ def read_http(sock):
 
     return response_line, headers, body
 
+
 class _TestBase(LimitedTestCase):
     def setUp(self):
         super(_TestBase, self).setUp()
@@ -174,7 +175,6 @@ class _TestBase(LimitedTestCase):
         eventlet.sleep(0) # give previous server a chance to start
         if self.killer:
             greenthread.kill(self.killer)
-            eventlet.sleep(0) # give killer a chance to kill
 
         new_kwargs = dict(max_size=128,
                           log=self.logfile,
@@ -192,6 +192,7 @@ class _TestBase(LimitedTestCase):
     def set_site(self):
         raise NotImplementedError
 
+
 class TestHttpd(_TestBase):
     def set_site(self):
         self.site = Site()
@@ -371,6 +372,7 @@ class TestHttpd(_TestBase):
         # Require a CRLF to close the message body
         self.assertEqual(response, '\r\n')
 
+    @skip_if_no_ssl
     def test_012_ssl_server(self):
         def wsgi_app(environ, start_response):
             start_response('200 OK', {})
@@ -379,18 +381,19 @@ class TestHttpd(_TestBase):
         certificate_file = os.path.join(os.path.dirname(__file__), 'test_server.crt')
         private_key_file = os.path.join(os.path.dirname(__file__), 'test_server.key')
 
-        server_sock = eventlet.wrap_ssl(eventlet.listen(('localhost', 0)), 
-                                        certfile=certificate_file, 
+        server_sock = eventlet.wrap_ssl(eventlet.listen(('localhost', 0)),
+                                        certfile=certificate_file,
                                         keyfile=private_key_file,
                                         server_side=True)
         self.spawn_server(sock=server_sock, site=wsgi_app)
-    
+
         sock = eventlet.connect(('localhost', self.port))
         sock = eventlet.wrap_ssl(sock)
         sock.write('POST /foo HTTP/1.1\r\nHost: localhost\r\nConnection: close\r\nContent-length:3\r\n\r\nabc')
         result = sock.read(8192)
         self.assertEquals(result[-3:], 'abc')
-        
+
+    @skip_if_no_ssl
     def test_013_empty_return(self):
         def wsgi_app(environ, start_response):
             start_response("200 OK", [])
@@ -398,8 +401,8 @@ class TestHttpd(_TestBase):
 
         certificate_file = os.path.join(os.path.dirname(__file__), 'test_server.crt')
         private_key_file = os.path.join(os.path.dirname(__file__), 'test_server.key')
-        server_sock = eventlet.wrap_ssl(eventlet.listen(('localhost', 0)), 
-                                        certfile=certificate_file, 
+        server_sock = eventlet.wrap_ssl(eventlet.listen(('localhost', 0)),
+                                        certfile=certificate_file,
                                         keyfile=private_key_file,
                                         server_side=True)
         self.spawn_server(sock=server_sock, site=wsgi_app)
@@ -488,6 +491,7 @@ class TestHttpd(_TestBase):
         self.assertEquals(1, len([l for l in header_lines
                 if l.lower().startswith('content-length')]))
 
+    @skip_if_no_ssl
     def test_017_ssl_zeroreturnerror(self):
 
         def server(sock, site, log):
@@ -508,8 +512,8 @@ class TestHttpd(_TestBase):
         certificate_file = os.path.join(os.path.dirname(__file__), 'test_server.crt')
         private_key_file = os.path.join(os.path.dirname(__file__), 'test_server.key')
 
-        sock = eventlet.wrap_ssl(eventlet.listen(('localhost', 0)), 
-                                        certfile=certificate_file, 
+        sock = eventlet.wrap_ssl(eventlet.listen(('localhost', 0)),
+                                        certfile=certificate_file,
                                         keyfile=private_key_file,
                                         server_side=True)
         server_coro = eventlet.spawn(server, sock, wsgi_app, self.logfile)
@@ -522,7 +526,7 @@ class TestHttpd(_TestBase):
 
         success = server_coro.wait()
         self.assert_(success)
-        
+
     def test_018_http_10_keepalive(self):
         # verify that if an http/1.0 client sends connection: keep-alive
         # that we don't close the connection
@@ -532,7 +536,7 @@ class TestHttpd(_TestBase):
         fd = sock.makefile('w')
         fd.write('GET / HTTP/1.0\r\nHost: localhost\r\nConnection: keep-alive\r\n\r\n')
         fd.flush()
-        
+
         response_line, headers, body = read_http(sock)
         self.assert_('connection' in headers)
         self.assertEqual('keep-alive', headers['connection'])
@@ -542,7 +546,7 @@ class TestHttpd(_TestBase):
         response_line, headers, body = read_http(sock)
         self.assert_('connection' in headers)
         self.assertEqual('keep-alive', headers['connection'])
-                     
+
     def test_019_fieldstorage_compat(self):
         def use_fieldstorage(environ, start_response):
             import cgi
@@ -550,7 +554,7 @@ class TestHttpd(_TestBase):
                                   environ=environ)
             start_response('200 OK', [('Content-type', 'text/plain')])
             return ['hello!']
-                             
+
         self.site.application = use_fieldstorage
         sock = eventlet.connect(
             ('localhost', self.port))
@@ -571,17 +575,17 @@ class TestHttpd(_TestBase):
         sock.recv(1024)
         sock.close()
         self.assert_('1.2.3.4,5.6.7.8,127.0.0.1' in self.logfile.getvalue())
-        
+
         # turning off the option should work too
         self.logfile = StringIO()
         self.spawn_server(log_x_forwarded_for=False)
-            
+
         sock = eventlet.connect(('localhost', self.port))
         sock.sendall('GET / HTTP/1.1\r\nHost: localhost\r\nX-Forwarded-For: 1.2.3.4, 5.6.7.8\r\n\r\n')
         sock.recv(1024)
         sock.close()
         self.assert_('1.2.3.4' not in self.logfile.getvalue())
-        self.assert_('5.6.7.8' not in self.logfile.getvalue())        
+        self.assert_('5.6.7.8' not in self.logfile.getvalue())
         self.assert_('127.0.0.1' in self.logfile.getvalue())
 
     def test_socket_remains_open(self):
@@ -623,8 +627,8 @@ class TestHttpd(_TestBase):
             for environ_var in ['wsgi.version', 'wsgi.url_scheme',
                 'wsgi.input', 'wsgi.errors', 'wsgi.multithread',
                 'wsgi.multiprocess', 'wsgi.run_once', 'REQUEST_METHOD',
-                'SCRIPT_NAME', 'PATH_INFO', 'QUERY_STRING', 'CONTENT_TYPE',
-                'CONTENT_LENGTH', 'SERVER_NAME', 'SERVER_PORT', 
+                'SCRIPT_NAME', 'RAW_PATH_INFO', 'PATH_INFO', 'QUERY_STRING',
+                'CONTENT_TYPE', 'CONTENT_LENGTH', 'SERVER_NAME', 'SERVER_PORT',
                 'SERVER_PROTOCOL']:
                 environ[environ_var] = None
             start_response('200 OK', [('Content-type', 'text/plain')])
@@ -646,7 +650,7 @@ class TestHttpd(_TestBase):
         from eventlet import greenpool
         p = greenpool.GreenPool(5)
         self.spawn_server(custom_pool=p)
-            
+
         # this stuff is copied from test_001_server, could be better factored
         sock = eventlet.connect(
             ('localhost', self.port))
@@ -742,18 +746,18 @@ class TestHttpd(_TestBase):
 
     def test_close_chunked_with_1_0_client(self):
         # verify that if we return a generator from our app
-        # and we're not speaking with a 1.1 client, that we 
+        # and we're not speaking with a 1.1 client, that we
         # close the connection
         self.site.application = chunked_app
         sock = eventlet.connect(('localhost', self.port))
 
         sock.sendall('GET / HTTP/1.0\r\nHost: localhost\r\nConnection: keep-alive\r\n\r\n')
-        
+
         response_line, headers, body = read_http(sock)
         self.assertEqual(headers['connection'], 'close')
         self.assertNotEqual(headers.get('transfer-encoding'), 'chunked')
         self.assertEquals(body, "thisischunked")
-        
+
     def test_026_http_10_nokeepalive(self):
         # verify that if an http/1.0 client sends connection: keep-alive
         # and the server doesn't accept keep-alives, we close the connection
@@ -782,6 +786,7 @@ class TestHttpd(_TestBase):
         fd.flush()
         read_http(sock)
 
+    @skip_if_no_ssl
     def test_028_ssl_handshake_errors(self):
         errored = [False]
         def server(sock):
@@ -793,15 +798,15 @@ class TestHttpd(_TestBase):
             except Exception, e:
                 errored[0] = 'SSL handshake error raised exception %s.' % e
         for data in ('', 'GET /non-ssl-request HTTP/1.0\r\n\r\n'):
-            srv_sock = eventlet.wrap_ssl(eventlet.listen(('localhost', 0)), 
-                                        certfile=certificate_file, 
+            srv_sock = eventlet.wrap_ssl(eventlet.listen(('localhost', 0)),
+                                        certfile=certificate_file,
                                         keyfile=private_key_file,
                                         server_side=True)
             port = srv_sock.getsockname()[1]
             g = eventlet.spawn_n(server, srv_sock)
             client = eventlet.connect(('localhost', port))
             if data: # send non-ssl request
-                client.sendall(data) 
+                client.sendall(data)
             else: # close sock prematurely
                 client.close()
             eventlet.sleep(0) # let context switch back to server
@@ -872,6 +877,30 @@ class TestHttpd(_TestBase):
         self.assertEquals(posthook1_count[0], 26)
         self.assertEquals(posthook2_count[0], 25)
 
+    def test_030_reject_long_header_lines(self):
+        sock = eventlet.connect(('localhost', self.port))
+        request = 'GET / HTTP/1.0\r\nHost: localhost\r\nLong: %s\r\n\r\n' % \
+            ('a' * 10000)
+        fd = sock.makefile('rw')
+        fd.write(request)
+        fd.flush()
+        response_line, headers, body = read_http(sock)
+        self.assertEquals(response_line,
+            'HTTP/1.0 400 Header Line Too Long\r\n')
+        fd.close()
+
+    def test_031_reject_large_headers(self):
+        sock = eventlet.connect(('localhost', self.port))
+        headers = 'Name: Value\r\n' * 5050
+        request = 'GET / HTTP/1.0\r\nHost: localhost\r\n%s\r\n\r\n' % headers
+        fd = sock.makefile('rw')
+        fd.write(request)
+        fd.flush()
+        response_line, headers, body = read_http(sock)
+        self.assertEquals(response_line,
+            'HTTP/1.0 400 Headers Too Large\r\n')
+        fd.close()
+
     def test_zero_length_chunked_response(self):
         def zero_chunked_app(env, start_response):
             start_response('200 OK', [('Content-type', 'text/plain')])
@@ -896,6 +925,23 @@ class TestHttpd(_TestBase):
         # (one terminates the chunk, one terminates the body)
         self.assertEqual(response, ['0', '', ''])
 
+    def test_configurable_url_length_limit(self):
+        self.spawn_server(url_length_limit=20000)
+        sock = eventlet.connect(
+            ('localhost', self.port))
+        path = 'x' * 15000
+        request = 'GET /%s HTTP/1.0\r\nHost: localhost\r\n\r\n' % path
+        fd = sock.makefile('rw')
+        fd.write(request)
+        fd.flush()
+        result = fd.readline()
+        if result:
+            # windows closes the socket before the data is flushed,
+            # so we never get anything back
+            status = result.split(' ')[1]
+            self.assertEqual(status, '200')
+        fd.close()
+
     def test_aborted_chunked_post(self):
         read_content = event.Event()
         blew_up = [False]
@@ -952,6 +998,98 @@ class TestHttpd(_TestBase):
         self.assertEqual(headers['connection'], 'close')
         self.assert_('unicode' in body)
 
+    def test_path_info_decoding(self):
+        def wsgi_app(environ, start_response):
+            start_response("200 OK", [])
+            yield "decoded: %s" % environ['PATH_INFO']
+            yield "raw: %s" % environ['RAW_PATH_INFO']
+        self.site.application = wsgi_app
+        sock = eventlet.connect(('localhost', self.port))
+        fd = sock.makefile('rw')
+        fd.write('GET /a*b@%40%233 HTTP/1.1\r\nHost: localhost\r\nConnection: '\
+                'close\r\n\r\n')
+        fd.flush()
+        response_line, headers, body = read_http(sock)
+        self.assert_(response_line.startswith('HTTP/1.1 200'))
+        self.assert_('decoded: /a*b@@#3' in body)
+        self.assert_('raw: /a*b@%40%233' in body)
+
+    def test_ipv6(self):
+        try:
+            sock = eventlet.listen(('::1', 0), family=socket.AF_INET6)
+        except (socket.gaierror, socket.error):  # probably no ipv6
+            return
+        log = StringIO()
+        # first thing the server does is try to log the IP it's bound to
+        def run_server():
+            try:
+                server = wsgi.server(sock=sock, log=log, site=Site())
+            except ValueError:
+                log.write('broked')
+        eventlet.spawn_n(run_server)
+        logval = log.getvalue()
+        while not logval:
+            eventlet.sleep(0.0)
+            logval = log.getvalue()
+        if 'broked' in logval:
+            self.fail('WSGI server raised exception with ipv6 socket')
+
+    def test_debug(self):
+        self.spawn_server(debug=False)
+        def crasher(env, start_response):
+            raise RuntimeError("intentional crash")
+        self.site.application = crasher
+
+        sock = eventlet.connect(('localhost', self.port))
+        fd = sock.makefile('w')
+        fd.write('GET / HTTP/1.1\r\nHost: localhost\r\n\r\n')
+        fd.flush()
+        response_line, headers, body = read_http(sock)
+        self.assert_(response_line.startswith('HTTP/1.1 500 Internal Server Error'))
+        self.assertEqual(body, '')
+        self.assertEqual(headers['connection'], 'close')
+        self.assert_('transfer-encoding' not in headers)
+
+        # verify traceback when debugging enabled
+        self.spawn_server(debug=True)
+        self.site.application = crasher
+        sock = eventlet.connect(('localhost', self.port))
+        fd = sock.makefile('w')
+        fd.write('GET / HTTP/1.1\r\nHost: localhost\r\n\r\n')
+        fd.flush()
+        response_line, headers, body = read_http(sock)
+        self.assert_(response_line.startswith('HTTP/1.1 500 Internal Server Error'))
+        self.assert_('intentional crash' in body)
+        self.assert_('RuntimeError' in body)
+        self.assert_('Traceback' in body)
+        self.assertEqual(headers['connection'], 'close')
+        self.assert_('transfer-encoding' not in headers)
+
+    def test_client_disconnect(self):
+        """Issue #95 Server must handle disconnect from client in the middle of response
+        """
+        def long_response(environ, start_response):
+            start_response('200 OK', [('Content-Length', '9876')])
+            yield 'a' * 9876
+
+        server_sock = eventlet.listen(('localhost', 0))
+        self.port = server_sock.getsockname()[1]
+        server = wsgi.Server(server_sock, server_sock.getsockname(), long_response,
+                             log=self.logfile)
+
+        def make_request():
+            sock = eventlet.connect(server_sock.getsockname())
+            sock.send('GET / HTTP/1.1\r\nHost: localhost\r\n\r\n')
+            sock.close()
+
+        request_thread = eventlet.spawn(make_request)
+        server_conn = server_sock.accept()
+        # Next line must not raise IOError -32 Broken pipe
+        server.process_request(server_conn)
+        request_thread.wait()
+        server_sock.close()
+
+
 def read_headers(sock):
     fd = sock.makefile()
     try:
@@ -1138,7 +1276,6 @@ class TestChunkedInput(_TestBase):
 
         assert not got_signal, "caught alarm signal. infinite loop detected."
 
-        
 
 if __name__ == '__main__':
     main()
